@@ -44,6 +44,7 @@
 #include "forward_decls.h"
 #include "item.h"
 #include "itempool.h"
+#include "moduleproviderinfo.h"
 #include <logging/logger.h>
 #include <tools/filetime.h>
 #include <tools/qttools.h>
@@ -57,6 +58,7 @@
 
 #include <map>
 #include <memory>
+#include <optional>
 #include <unordered_map>
 #include <utility>
 #include <vector>
@@ -71,8 +73,11 @@ namespace Internal {
 class Evaluator;
 class Item;
 class ItemReader;
+class ModuleProviderLoader;
+class ProbesResolver;
 class ProgressObserver;
 class QualifiedId;
+class SearchPathsManager;
 
 using ModulePropertiesPerGroup = std::unordered_map<const Item *, QualifiedIdSet>;
 
@@ -104,8 +109,9 @@ struct ModuleLoaderResult
 
     std::shared_ptr<ItemPool> itemPool;
     Item *root;
-    QHash<Item *, ProductInfo> productInfos;
+    std::unordered_map<Item *, ProductInfo> productInfos;
     std::vector<ProbeConstPtr> projectProbes;
+    StoredModuleProviderInfo storedModuleProviderInfo;
     Set<QString> qbsFiles;
     QVariantMap profileConfigs;
 };
@@ -128,11 +134,14 @@ public:
     void setOldProductProbes(const QHash<QString, std::vector<ProbeConstPtr>> &oldProbes);
     void setLastResolveTime(const FileTime &time) { m_lastResolveTime = time; }
     void setStoredProfiles(const QVariantMap &profiles);
+    void setStoredModuleProviderInfo(const StoredModuleProviderInfo &moduleProviderInfo);
     Evaluator *evaluator() const { return m_evaluator; }
 
     ModuleLoaderResult load(const SetupProjectParameters &parameters);
 
 private:
+    friend class ModuleProviderLoader;
+    friend class ProbesResolver;
     class ProductSortByDependencies;
 
     class ContextBase
@@ -149,7 +158,7 @@ private:
 
     class ProjectContext;
 
-    typedef std::vector<ModuleLoaderResult::ProductInfo::Dependency> ProductDependencies;
+    using ProductDependencies = std::vector<ModuleLoaderResult::ProductInfo::Dependency>;
 
     // This is the data we need to store at the point where a dependency is deferred
     // in order to properly resolve the dependency in pass 2.
@@ -172,7 +181,7 @@ private:
     class ProductContext : public ContextBase
     {
     public:
-        ProjectContext *project;
+        ProjectContext *project = nullptr;
         ModuleLoaderResult::ProductInfo info;
         QString profileName;
         QString multiplexConfigurationId;
@@ -181,6 +190,8 @@ private:
         std::unordered_map<const Item *, std::vector<ErrorInfo>> unknownProfilePropertyErrors;
         QStringList searchPaths;
 
+        std::optional<QVariantMap> theModuleProviderConfig;
+
         // The key corresponds to DeferredDependsContext.exportingProductItem, which is the
         // only value from that data structure that we still need here.
         std::unordered_map<Item *, std::vector<Item *>> deferredDependsItems;
@@ -193,8 +204,8 @@ private:
     class ProjectContext : public ContextBase
     {
     public:
-        TopLevelProjectContext *topLevelProject;
-        ModuleLoaderResult *result;
+        TopLevelProjectContext *topLevelProject = nullptr;
+        ModuleLoaderResult *result = nullptr;
         std::vector<ProductContext> products;
         std::vector<QStringList> searchPathsStack;
     };
@@ -210,7 +221,7 @@ private:
     {
         Q_DISABLE_COPY(TopLevelProjectContext)
     public:
-        TopLevelProjectContext() {}
+        TopLevelProjectContext() = default;
         ~TopLevelProjectContext() { qDeleteAll(projects); }
 
         std::vector<ProjectContext *> projects;
@@ -224,7 +235,7 @@ private:
     public:
         ProductContext *product = nullptr;
         Item *exportingProductItem = nullptr;
-        ProductDependencies *productDependencies;
+        ProductDependencies *productDependencies = nullptr;
     };
 
     void handleTopLevelProject(ModuleLoaderResult *loadResult, Item *projectItem,
@@ -233,8 +244,8 @@ private:
             TopLevelProjectContext *topLevelProjectContext, Item *projectItem,
             const Set<QString> &referencedFilePaths);
 
-    typedef std::vector<VariantValuePtr> MultiplexRow;
-    typedef std::vector<MultiplexRow> MultiplexTable;
+    using MultiplexRow = std::vector<VariantValuePtr>;
+    using MultiplexTable = std::vector<MultiplexRow>;
 
     struct MultiplexInfo
     {
@@ -244,6 +255,7 @@ private:
         VariantValuePtr multiplexedType;
 
         QString toIdString(size_t row) const;
+        static QVariantMap multiplexIdToVariantMap(const QString &multiplexId);
     };
 
     void dump(const MultiplexInfo &mpi);
@@ -297,20 +309,24 @@ private:
     QVariantMap extractParameters(Item *dependsItem) const;
     Item *moduleInstanceItem(Item *containerItem, const QualifiedId &moduleName);
     static ProductModuleInfo *productModule(ProductContext *productContext, const QString &name,
-                                            const QString &multiplexId);
+                                            const QString &multiplexId, bool &productNameMatch);
     static ProductContext *product(ProjectContext *projectContext, const QString &name);
     static ProductContext *product(TopLevelProjectContext *tlpContext, const QString &name);
+
+    enum class FallbackMode { Enabled, Disabled };
     Item *loadModule(ProductContext *productContext, Item *exportingProductItem, Item *item,
             const CodeLocation &dependsItemLocation, const QString &moduleId,
-            const QualifiedId &moduleName, const QString &multiplexId, bool isRequired,
-                     bool *isProductDependency, QVariantMap *defaultParameters);
+            const QualifiedId &moduleName, const QString &multiplexId, FallbackMode fallbackMode,
+            bool isRequired, bool *isProductDependency, QVariantMap *defaultParameters);
     Item *searchAndLoadModuleFile(ProductContext *productContext,
             const CodeLocation &dependsItemLocation, const QualifiedId &moduleName,
-            bool isRequired, Item *moduleInstance);
-    Item *loadModuleFile(ProductContext *productContext, const QString &fullModuleName,
-            bool isBaseModule, const QString &filePath, bool *triedToLoad, Item *moduleInstance);
-    Item *getModulePrototype(ProductContext *productContext, const QString &fullModuleName,
-                             const QString &filePath, bool *triedToLoad);
+            FallbackMode fallbackMode, bool isRequired, Item *moduleInstance);
+    QStringList &getModuleFileNames(const QString &dirPath);
+    std::pair<Item *, bool> loadModuleFile(
+            ProductContext *productContext, const QString &fullModuleName,
+            const QString &filePath, Item *moduleInstance);
+    std::pair<Item *, bool> getModulePrototype(ProductContext *productContext,
+            const QString &fullModuleName, const QString &filePath);
     Item::Module loadBaseModule(ProductContext *productContext, Item *item);
     void setupBaseModulePrototype(Item *prototype);
     template <typename T, typename F>
@@ -320,15 +336,15 @@ private:
             const QualifiedId &moduleName, ProductModuleInfo *productModuleInfo);
     void createChildInstances(Item *instance, Item *prototype,
                               QHash<Item *, Item *> *prototypeInstanceMap) const;
-    void resolveProbes(ProductContext *productContext, Item *item);
-    void resolveProbe(ProductContext *productContext, Item *parent, Item *probe);
     void checkCancelation() const;
     bool checkItemCondition(Item *item, Item *itemToDisable = nullptr);
-    QStringList readExtraSearchPaths(Item *item, bool *wasSet = 0);
+    QStringList readExtraSearchPaths(Item *item, bool *wasSet = nullptr);
     void copyProperties(const Item *sourceProject, Item *targetProject);
     Item *wrapInProjectIfNecessary(Item *item);
-    static QString findExistingModulePath(const QString &searchPath,
-            const QualifiedId &moduleName);
+    QString findExistingModulePath(const QString &searchPath, const QualifiedId &moduleName);
+    QStringList findExistingModulePaths(
+            const QStringList &searchPaths, const QualifiedId &moduleName);
+
     static void setScopeForDescendants(Item *item, Item *scope);
     void overrideItemProperties(Item *item, const QString &buildConfigKey,
                                 const QVariantMap &buildConfig);
@@ -341,19 +357,6 @@ private:
     void copyGroupsFromModulesToProduct(const ProductContext &productContext);
     void markModuleTargetGroups(Item *group, const Item::Module &module);
     bool checkExportItemCondition(Item *exportItem, const ProductContext &productContext);
-    ProbeConstPtr findOldProjectProbe(const QString &globalId, bool condition,
-                                      const QVariantMap &initialProperties,
-                                      const QString &sourceCode) const;
-    ProbeConstPtr findOldProductProbe(const QString &productName, bool condition,
-                                      const QVariantMap &initialProperties,
-                                      const QString &sourceCode) const;
-    ProbeConstPtr findCurrentProbe(const CodeLocation &location, bool condition,
-                                   const QVariantMap &initialProperties) const;
-
-    enum class CompareScript { No, Yes };
-    bool probeMatches(const ProbeConstPtr &probe, bool condition,
-                      const QVariantMap &initialProperties, const QString &configureScript,
-                      CompareScript compareScript) const;
 
     void printProfilingInfo();
     void handleProductError(const ErrorInfo &error, ProductContext *productContext);
@@ -384,9 +387,12 @@ private:
     ItemPool *m_pool;
     Logger &m_logger;
     ProgressObserver *m_progressObserver;
-    ItemReader *m_reader;
+    const std::unique_ptr<ItemReader> m_reader;
     Evaluator *m_evaluator;
+    const std::unique_ptr<ProbesResolver> m_probesResolver;
+    const std::unique_ptr<ModuleProviderLoader> m_moduleProviderLoader;
     QMap<QString, QStringList> m_moduleDirListCache;
+    QHash<std::pair<QString, QualifiedId>, std::optional<QString>> m_existingModulePathCache;
 
     // The keys are file paths, the values are module prototype items accompanied by a profile.
     std::unordered_map<QString, std::vector<std::pair<Item *, QString>>> m_modulePrototypes;
@@ -401,8 +407,8 @@ private:
 
     struct DependsChainEntry
     {
-        DependsChainEntry(const QualifiedId &name, const CodeLocation &location)
-            : name(name), location(location)
+        DependsChainEntry(QualifiedId name, const CodeLocation &location)
+            : name(std::move(name)), location(location)
         {
         }
 
@@ -413,10 +419,7 @@ private:
     class DependsChainManager;
     std::vector<DependsChainEntry> m_dependsChain;
 
-    QHash<QString, QList<ProbeConstPtr>> m_oldProjectProbes;
-    QHash<QString, std::vector<ProbeConstPtr>> m_oldProductProbes;
     FileTime m_lastResolveTime;
-    QHash<CodeLocation, QList<ProbeConstPtr>> m_currentProbes;
     QVariantMap m_storedProfiles;
     QVariantMap m_localProfiles;
     std::multimap<QString, const ProductContext *> m_productsByName;
@@ -430,16 +433,13 @@ private:
     Version m_qbsVersion;
     Item *m_tempScopeItem = nullptr;
 
-    qint64 m_elapsedTimeProbes;
-    qint64 m_elapsedTimePrepareProducts;
-    qint64 m_elapsedTimeProductDependencies;
-    qint64 m_elapsedTimeTransitiveDependencies;
-    qint64 m_elapsedTimeHandleProducts;
-    qint64 m_elapsedTimePropertyChecking;
-    quint64 m_probesEncountered;
-    quint64 m_probesRun;
-    quint64 m_probesCachedCurrent;
-    quint64 m_probesCachedOld;
+    qint64 m_elapsedTimeProbes = 0;
+    qint64 m_elapsedTimePrepareProducts = 0;
+    qint64 m_elapsedTimeProductDependencies = 0;
+    qint64 m_elapsedTimeModuleProviders = 0;
+    qint64 m_elapsedTimeTransitiveDependencies = 0;
+    qint64 m_elapsedTimeHandleProducts = 0;
+    qint64 m_elapsedTimePropertyChecking = 0;
     Set<QString> m_projectNamesUsedInOverrides;
     Set<QString> m_productNamesUsedInOverrides;
     Set<QString> m_disabledProjects;
